import logging
from datetime import datetime, timedelta
from ogd.games.JOURNALISM.features import PlayTime
from ogd.core.models import Event
from typing import Any, Final, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class LevelTime(PerLevelFeature):
    def __init__(self, params:GeneratorParameters):
        PerLevelFeature.__init__(self, params=params)
        self._begin_times    : List[datetime] = []
        self._complete_times : List[datetime] = []
        self._idle_time  : Optional[timedelta] = None
        self._total_time : Optional[timedelta] = None


    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["start_level", "complete_level"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["PlayTime"]

    def _validateEventCountIndex(self, event: Event):
        return self.CountIndex == 1
        


    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "start_level":
            self._begin_times.append(event.Timestamp)
        elif event.EventName == "complete_level":
            self._complete_times.append(event.Timestamp)
        else:
            Logger.Log(f"LevelTime received an event which was not a BEGIN or a COMPLETE!", logging.WARN)

    def _updateFromFeatureData(self, feature:FeatureData):
        IDLE_TIME_INDEX : Final[int] = 2 # Idle time should be at index 2 for the PlayTime feature
        if feature.FeatureType == "PlayTime":
            self._idle_time = feature.FeatureValues[IDLE_TIME_INDEX]

    def _getFeatureValues(self) -> List[Any]:
        if len(self._begin_times) < len(self._complete_times):
            Logger.Log(f"Player began level {self.CountIndex} {len(self._begin_times)} times but completed it {len(self._complete_times)}.", logging.DEBUG)
        _num_plays = min(len(self._begin_times), len(self._complete_times))
        _diffs = [(self._complete_times[i] - self._begin_times[i]).total_seconds() for i in range(_num_plays)]
        self._total_time = sum(_diffs)
        total_time = self._total_time
        play_time = total_time - self._idle_time.total_seconds()
        return [play_time, total_time, self._idle_time]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Total Time", "Idle Time"]
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.SESSION]





from typing import Any, Dict, List

from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class LeftHandMovesCount(SessionFeature):

    def __init__(self, params: GeneratorParameters):
        self._left_hand_move_count = 0
        self._last_hand_action = None
        super().__init__(params=params)

    @classmethod
    def _getEventDependencies(cls, mode: ExtractionMode) -> List[str]:
        return ["grab_tool_slider", "release_tool_slider", "click_new_game", "click_reset_sim", 
                "grab_tablet", "release_tablet", "grab_workstation_handle", "release_workstation_handle",
                "click_rotate_graph_cw", "click_rotate_graph_ccw", "grab_graph_ball", "release_graph_ball",
                "click_tool_toggle", "click_tool_increase", "click_tool_decrease",
                "click_view_settings", "click_toggle_setting", 
                "click_sandbox_mode", "click_lab_mode", "click_lab_scroll_up", "click_lab_scroll_down"]

    @classmethod
    def _getFeatureDependencies(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _extractFromEvent(self, event: Event) -> None:
        hand = event.EventData.get("hand")
        if hand == "LEFT":
            action = event.EventName.split("_")[0]  # Extract action (grab, release, click, etc.)
            if action in ["grab", "click"] or (action == "release" and self._last_hand_action != "grab"):
                self._left_hand_move_count += 1
            self._last_hand_action = action

    def _extractFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._left_hand_move_count]



# import standard libraries
import itertools
import json
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional
# import local files
from . import features
from ogd.games import THERMOVR
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.Feature import Feature
from ogd.games.THERMOVR.detectors import *
from ogd.games.THERMOVR.features import *
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.utils import loadJSONFile
from ogd.core.utils.Logger import Logger

EXPORT_PATH = "games/THERMOVR/DBExport.json"

## @class ThermoVRLoader
#  Extractor subclass for extracting features from ThermoVR game data.
class ThermoVRLoader(GeneratorLoader):

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the ThermoVRLoader class.
    def __init__(self, player_id:str, session_id:str, game_schema: GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        """Constructor for the ThermoVRLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)
        self._lab_map = {}
        data = None

        # Load ThermoVR jobs export and map job names to integer values
        _dbexport_path = Path(THERMOVR.__file__) if Path(THERMOVR.__file__).is_dir() else Path(THERMOVR.__file__).parent
        with open(_dbexport_path / "DBExport.json", "r") as file:
            export = json.load(file)

            # task_num = 1
            # for i, lab in enumerate(export["labs"], start=1):
            #     self._lab_map[lab["id"]] = i
            #     for task in lab["tasks"]:
            #         task_by_lab = lab["id"] + "_" + task["id"]
            #         self._task_map[task_by_lab] = task_num
            #         task_num += 1

        # Update level count
        self._game_schema._max_level = len(self._lab_map) - 1

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule():
        return features

    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Optional[Feature]:
        ret_val : Optional[Feature] = None
        # First run through aggregate features
        if extractor_params._count_index == None:
            match feature_type:
                case "PhasesReached":
                    ret_val = PhasesReached.PhasesReached(params=extractor_params)
                case "PlayMode":
                    ret_val = PlayMode.PlayMode(params=extractor_params)
                case "TaskCompleteCount":
                    ret_val = TaskCompleteCount.TaskCompleteCount(params=extractor_params)
                case "LabCompleteCount":
                    ret_val = LabCompleteCount.LabCompleteCount(params=extractor_params)
                case "LeftHandMoves":
                    ret_val = LeftHandMoves.LeftHandMoves(params=extractor_params)
                case "RightHandMoves":
                    ret_val = RightHandMoves.RightHandMoves(params=extractor_params)
                case "ToolNudgeCount":
                    ret_val = ToolNudgeCount.ToolNudgeCount(params=extractor_params)
                case "ToolSliderTime":
                    ret_val = ToolSliderTime.ToolSliderTime(params=extractor_params)
                case _:
                    Logger.Log(f"'{feature_type}' is not a valid aggregate feature type for ThermoVR.")
        # then run through per-count features.
        else:
            match feature_type:
                case _:
                    Logger.Log(f"'{feature_type}' is not a valid per-count feature type for ThermoVR.")
        return ret_val

    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Optional[Detector]:
        ret_val : Optional[Detector] = None
        match detector_type:
            case _:
                Logger.Log(f"'{detector_type}' is not a valid detector for ThermoVR.")
        return ret_val

    # @staticmethod
    # def GetThermoVRLabCount(db_export_path:Path=Path(".") / "ogd" / "games" / "THERMOVR"):
    #     db_export = loadJSONFile(filename="DBExport.json", path=db_export_path)
    #     return len(db_export.get("jobs", []))

    # @staticmethod
    # def GetThermoVRTaskCount(db_export_path:Path=Path(".") / "ogd" / "games" / "THERMOVR"):
    #     db_export = loadJSONFile(filename="DBExport.json", path=db_export_path)
    #     list_o_lists = [job.get('tasks', []) for job in db_export.get('jobs', [])]
    #     # jobs_to_task_cts = [f"{job.get('id')}: {len(job.get('tasks', []))}" for job in db_export.get('jobs', [])]
    #     # Logger.Log(f"Task counts by job:\n{jobs_to_task_cts}", logging.DEBUG)
    #     all_tasks    = list(itertools.chain.from_iterable(list_o_lists))
    #     return len(all_tasks)



# import standard libraries
from datetime import datetime
from typing import Any, List, Optional
# import local files
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import EventSource
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode

class LegacyDetector(Detector):
    """Dummy version of a detector, so that LegacyLoader can return something that's not None.
    """

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of events this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        return

    def _trigger_condition(self) -> bool:
        return False

    def _trigger_event(self) -> Optional[Event]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        ret_val : Event = Event(session_id="Not Implemented", app_id="Not Implemented", timestamp=datetime.now(),
                                event_name="CustomDetector", event_data={}, event_source=EventSource.GENERATED)
        return ret_val

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params, trigger_callback=lambda x : None)



import logging
from datetime import date, datetime, timedelta, timezone
from enum import IntEnum
from typing import Dict, List, Optional, Union

from ogd.core.utils import utils
class EventSource(IntEnum):
    GAME = 1
    GENERATED = 2

## @class Event
#  Completely dumb struct that enforces a particular structure for the data we get from a source.
#  Basically, whenever we fetch data, the TableSchema will be used to map columns to the required elements of an Event.
#  Then the extractors etc. can just access columns in a direct manner.
class Event:
    def __init__(self, session_id:str, app_id:str,     timestamp:datetime,
                 event_name:str, event_data:utils.map, event_source:EventSource,
                 app_version:Optional[str] = None,     app_branch:Optional[str] = None,
                 log_version:Optional[str] = None,     time_offset:Optional[timezone] = None,
                 user_id:Optional[str] = "",           user_data:Optional[utils.map] = {},
                 game_state:Optional[utils.map] = {},  event_sequence_index:Optional[int] = None):
        """Constructor for an Event object.

        :param session_id: An identifier for the session during which the event occurred.
        :type  session_id: int
        :param app_id: An identifier for the app that generated the event, typically the name of a game
        :type  app_id: str
        :param timestamp: The (local) time at which the event occurred.
        :type  timestamp: datetime
        :param event_name: The "type" of the event. e.g. begin game, end game, buy item, etc.
        :type  event_name: str
        :param event_data: A "blob" of all data specific to the event type, contents vary by game and event type.
        :type  event_data: Dict[str,Any]
        :param app_version: The version of the given game that created the event.
        :type  app_version: Optional[int], optional
        :param log_version: The version of the given game's logging code (may or may not correspond to game's versioning)
        :type  log_version: Optional[int], optional
        :param time_offset: [description], defaults to None
        :type  time_offset: Optional[int], optional
        :param user_id: Optional identifier for the specific user during whose session the event occurred. Defaults to None.
        :type  user_id: Optional[int], optional
        :param user_data: [description], defaults to None
        :type  user_data: Optional[Dict[str,Any]], optional
        :param game_state: [description], defaults to None
        :type  game_state: Optional[Dict[str,Any]], optional
        :param event_sequence_index: [description], defaults to None
        :type  event_sequence_index: Optional[int], optional
        """
        # TODO: event source, e.g. from game or from detector
        self.session_id           : str           = session_id
        self.app_id               : str           = app_id
        self.timestamp            : datetime      = timestamp
        self.event_name           : str           = event_name
        self.event_data           : utils.map     = event_data
        self.event_source         : EventSource   = event_source
        self.app_version          : str           = app_version if app_version is not None else "0"
        self.app_branch           : str           = app_branch  if app_branch  is not None else "main"
        self.log_version          : str           = log_version if log_version is not None else "0"
        self.time_offset          : Optional[timezone] = time_offset
        self.user_id              : Optional[str] = user_id
        self.user_data            : utils.map     = user_data if user_data is not None else {}
        self.game_state           : utils.map     = game_state if game_state is not None else {}
        self.event_sequence_index : Optional[int] = event_sequence_index

    def __str__(self):
        return f"session_id   : {self.session_id}\n"\
             + f"app_id       : {self.app_id}\n"\
             + f"timestamp    : {self.timestamp}\n"\
             + f"event_name   : {self.event_name}\n"\
             + f"event_data   : {self.event_data}\n"\
             + f"event_source : {self.event_source.name}\n"\
             + f"app_version  : {self.app_version}\n"\
             + f"app_branch   : {self.app_branch}\n"\
             + f"log_version  : {self.log_version}\n"\
             + f"offset       : {self.TimeOffsetString}\n"\
             + f"user_id      : {self.user_id}\n"\
             + f"user_data    : {self.user_data}\n"\
             + f"game_state   : {self.game_state}\n"\
             + f"index        : {self.event_sequence_index}\n"\

    def FallbackDefaults(self, app_id:Optional[str]=None, index:Optional[int]=None):
        if self.app_id == None and app_id != None:
            self.app_id = app_id
        if self.event_sequence_index == None:
            self.event_sequence_index = index

    @staticmethod
    def FromJSON(json_data:Dict):
        return Event(
            session_id  =json_data.get("session_id", "SESSION ID NOT FOUND"),
            app_id      =json_data.get("app_id", "APP ID NOT FOUND"),
            timestamp   =json_data.get("client_time", "CLIENT TIME NOT FOUND"),
            event_name  =json_data.get("event_name", "EVENT NAME NOT FOUND"),
            event_data  =json_data.get("event_data", "EVENT DATA NOT FOUND"),
            event_source=EventSource.GAME,
            app_version =json_data.get("app_version", None),
            app_branch  =json_data.get("app_branch", None),
            log_version =json_data.get("log_version", None),
            time_offset =None,
            user_id     =json_data.get("user_id", None),
            user_data   =json_data.get("user_data", None),
            game_state  =json_data.get("game_state", None),
            event_sequence_index=json_data.get("event_sequence_index", json_data).get("session_n", None)
        )

    @staticmethod
    def CompareVersions(a:str, b:str, version_separator='.') -> int:
        a_parts : Optional[List[int]]
        b_parts : Optional[List[int]]
        try:
            a_parts = [int(i) for i in a.split(version_separator)]
        except ValueError:
            a_parts = None
        try:
            b_parts = [int(i) for i in b.split(version_separator)]
        except ValueError:
            b_parts = None

        if a_parts is not None and b_parts is not None:
            for i in range(0, min(len(a_parts), len(b_parts))):
                if a_parts[i] < b_parts[i]:
                    return -1
                elif a_parts[i] > b_parts[i]:
                    return 1
            if len(a_parts) < len(b_parts):
                return -1
            elif len(a_parts) > len(b_parts):
                return 1
            else:
                return 0
        else:
            # try to do some sort of sane handling in case we got null values for a version
            if a_parts is None and b_parts is None:
                utils.Logger.Log(f"Got invalid values of {a} & {b} for versions a & b!", logging.ERROR)
                return 0
            elif a_parts is None:
                utils.Logger.Log(f"Got invalid value of {a} for version a!", logging.ERROR)
                return 1
            elif b_parts is None:
                utils.Logger.Log(f"Got invalid value of {b} for version b!", logging.ERROR)
                return -1
        return 0 # should never reach here; just putting this here to satisfy linter

    @staticmethod
    def ColumnNames() -> List[str]:
        return ["session_id",  "app_id",       "timestamp",   "event_name",
                "event_data",  "event_source", "app_version", "app_branch",
                "log_version", "offset",        "user_id",    "user_data",
                "game_state",  "index"]

    def ColumnValues(self) -> List[Union[str, datetime, timezone, utils.map, int, None]]:
        """A list of all values for the row, in order they appear in the `ColumnNames` function.

        .. todo:: Technically, this should be string representations of each, but we're technically not enforcing that yet.

        :return: The list of values.
        :rtype: List[Union[str, datetime, timezone, utils.map, int, None]]
        """
        return [self.session_id,  self.app_id,             self.timestamp,   self.event_name,
                self.event_data,  self.event_source.name,  self.app_version, self.app_branch,
                self.log_version, self.TimeOffsetString,   self.user_id,     self.user_data,
                self.game_state,  self.event_sequence_index]

    @property
    def SessionID(self) -> str:
        """The Session ID of the session that generated the Event

        Generally, this will be a numeric string.
        Every session ID is unique (with high probability) from all other sessions.

        :return: The Session ID of the session that generated the Event
        :rtype: str
        """
        return self.session_id

    @property
    def AppID(self) -> str:
        """The Application ID of the game that generated the Event

        Generally, this will be the game's name, or some abbreviation of the name.

        :return: The Application ID of the game that generated the Event
        :rtype: str
        """
        return self.app_id

    @property
    def Timestamp(self) -> datetime:
        """A UTC timestamp of the moment at which the game client sent the Event

        The timestamp is based on the GMT timezone, in keeping with UTC standards.
        Some legacy games may provide the time based on a local time zone, rather than GMT.

        :return: A UTC timestamp of the moment at which the game client sent the event
        :rtype: datetime
        """
        return self.timestamp

    @property
    def TimeOffset(self) -> Optional[timezone]:
        """A timedelta for the offset from GMT to the local time zone of the game client that sent the Event

        Some legacy games do not include an offset, and instead log the Timestamp based on the local time zone.

        :return: A timedelta for the offset from GMT to the local time zone of the game client that sent the Event
        :rtype: Optional[timedelta]
        """
        return self.time_offset

    @property
    def TimeOffsetString(self) -> Optional[str]:
        """A string representation of the offset from GMT to the local time zone of the game client that sent the Event

        Some legacy games do not include an offset, and instead log the Timestamp based on the local time zone.

        :return: A timedelta for the offset from GMT to the local time zone of the game client that sent the Event
        :rtype: Optional[timedelta]
        """
        return self.time_offset.tzname(None) if self.time_offset is not None else None

    @property
    def EventName(self) -> str:
        """The name of the specific type of event that occurred

        For some legacy games, the names in this column have a format of CUSTOM.1, CUSTOM.2, etc.
        For these games, the actual human-readable event names for these events are stored in the EventData column.
        Please see individual game logging documentation for details.

        :return: The name of the specific type of event that occurred
        :rtype: str
        """
        return self.event_name

    @property
    def EventData(self) -> utils.map:
        """A dictionary containing data specific to Events of this type.

        For details, see the documentation in the given game's README.md, included with all datasets.
        Alternately, review the {GAME_NAME}.json file for the given game.

        :return: A dictionary containing data specific to Events of this type
        :rtype: Dict[str, Any]
        """
        return self.event_data

    @property
    def EventSource(self) -> EventSource:
        """An enum indicating whether the event was generated directly by the game, or calculated by a post-hoc detector.

        :return: An enum indicating whether the event was generated directly by the game, or calculated by a post-hoc detector
        :rtype: EventSource
        """
        return self.event_source

    @property
    def AppVersion(self) -> str:
        """The semantic versioning string for the game that generated this Event.

        Some legacy games may use a single integer or a string similar to AppID in this column.

        :return: The semantic versioning string for the game that generated this Event
        :rtype: str
        """
        return self.app_version

    @property
    def AppBranch(self) -> str:
        """The name of the branch of a game version that generated this Event.

        The branch name is typically used for cases where multiple experimental versions of a game are deployed in parallel;
        most events will simply have a branch of "main" or "master."

        :return: The name of the branch of a game version that generated this Event
        :rtype: str
        """
        return self.app_branch

    @property
    def LogVersion(self) -> str:
        """The version of the logging schema implemented in the game that generated the Event

        For most games, this is a single integer; however, semantic versioning is valid for this column as well.

        :return: The version of the logging schema implemented in the game that generated the Event
        :rtype: str
        """
        return self.log_version

    @property
    def UserID(self) -> Optional[str]:
        """A persistent ID for a given user, identifying the individual across multiple gameplay sessions

        This identifier is only included by games with a mechanism for individuals to resume play in a new session.

        :return: A persistent ID for a given user, identifying the individual across multiple gameplay sessions
        :rtype: Optional[str]
        """
        return self.user_id

    @property
    def UserData(self) -> utils.map:
        """A dictionary containing any user-specific data tracked across gameplay sessions or individual games.

        :return: A dictionary containing any user-specific data tracked across gameplay sessions or individual games
        :rtype: Dict[str, Any]
        """
        return self.user_data

    @property
    def GameState(self) -> utils.map:
        """A dictionary containing any game-specific data that is defined across all event types in the given game.

        This column typically includes data that offers context to a given Event's data in the EventData column.
        For example, this column would typically include a level number or quest name for whatever level/quest the user was playing when the Event occurred.

        :return: A dictionary containing any game-specific data that is defined across all event types in the given game
        :rtype: Dict[str, Any]
        """
        return self.game_state

    @property
    def EventSequenceIndex(self) -> Optional[int]:
        """A strictly-increasing counter indicating the order of events in a session.

        The first event in a session has EventSequenceIndex == 0, the next has index == 1, etc.

        :return: A strictly-increasing counter indicating the order of events in a session
        :rtype: int
        """
        return self.event_sequence_index



# import libraries
import logging
from typing import Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.Event import Event

class PerJobFeature(PerCountFeature):
    def __init__(self, params:GeneratorParameters, job_map:dict):
        super().__init__(params=params,)
        self._job_map = job_map

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _validateEventCountIndex(self, event:Event):
        ret_val : bool = False

        job_name = event.GameState.get('job_name', event.EventData.get('job_name', "JOB NAME NOT FOUND"))
        if job_name is not None:
            if job_name in self._job_map and self._job_map[job_name] == self.CountIndex:
                ret_val = True
        else:
            self.WarningMessage(f"Got invalid job_name data in {type(self).__name__}")

        return ret_val

    # *** Optionally override public functions. ***

    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



"""# import libraries
from datetime import datetime, timedelta
import logging, warnings
from builtins import float
from typing import Any, Final, List, Optional, Dict
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class ActiveTime(Feature):
    IDLE_LEVEL: Final[int] = 30

    def __init__(self, params: GeneratorParameters, active_threads: Optional[float] = None):
        super().__init__(params=params)
        self._Idle_time: float = 0
        self.active_level: float = active_threads if active_threads else ActiveTime.IDLE_LEVEL
        self._sess_duration: Optional[timedelta] = None
        self._client_start_time: Optional[datetime] = None
        self._client_end_time: Optional[datetime] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["session_start", "pause_game", "unpause_game"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        if event.EventName == "session_start":
            self._client_start_time = event.Timestamp
        elif event.EventName == "pause_game":
            self._client_end_time = event.Timestamp
        elif event.EventName == "unpause_game":
            self._client_start_time = event.Timestamp

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if self._client_start_time and self._client_end_time:
            return [self._client_end_time - self._client_start_time - timedelta(seconds=self._Idle_time)]
        else:
            return ["No events"]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"
"""


# import libraries
from datetime import datetime, timedelta
from typing import Any, Final, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event, EventSource
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class ActiveTime(Feature):
    DEFAULT_IDLE_THRESHOLD: Final[timedelta] = timedelta(seconds=30)

    def __init__(self, params: GeneratorParameters, idle_threshold: Optional[int]):
        super().__init__(params=params)
        self.IDLE_THRESHOLD     : Final[timedelta] = timedelta(seconds=idle_threshold) if idle_threshold is not None else ActiveTime.DEFAULT_IDLE_THRESHOLD
        self.max_idle           : timedelta = timedelta(0)
        self.previous_time      : Optional[datetime] = None
        self.idle_time          : timedelta = timedelta(0)
        self.total_session_time : timedelta = timedelta(0)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["session_start", "pause_game", "unpause_game", "all_events"]

    def Subfeatures(self) -> List[str]:
        return ["Total", "Seconds", "ActiveSeconds", "Idle", "IdleSeconds", "MaxIdle"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        if event.EventSource == EventSource.GAME:
            if self.previous_time is not None:
                event_duration = event.Timestamp - self.previous_time
                self.total_session_time += event_duration
                if event_duration > self.IDLE_THRESHOLD:
                    self.idle_time += event_duration
                    if event_duration > self.max_idle:
                        self.max_idle = event_duration
            self.previous_time = event.Timestamp

        if event.EventName == "session_start":
            self.previous_time = event.Timestamp
        elif event.EventName in ["pause_game", "unpause_game"]:
            self.previous_time = event.Timestamp

    def _updateFromFeatureData(self, feature: FeatureData):
        pass

    def _getFeatureValues(self) -> List[Any]:
        active_time = self.total_session_time - self.idle_time
        return [
            active_time,
            self.total_session_time,
            self.total_session_time.total_seconds(),
            active_time.total_seconds(),
            self.idle_time,
            self.idle_time.total_seconds(),
            self.max_idle
        ]

    # *** Optionally override public functions ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



# import libraries
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class OverallPercentOffsetMoves(SessionFeature):
    def __init__(self, params:GeneratorParameters):
        SessionFeature.__init__(self, params=params)
        self._offset_count = 0
        self._move_count = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.1", "CUSTOM.2"]
        # return ["SLIDER_MOVE_RELEASE", "ARROW_MOVE_RELEASE"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        self._move_count += 1
        if event.EventData["slider"].upper() == "OFFSET":
            self._offset_count += 1

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if self._move_count > 0:
            return [self._offset_count / self._move_count * 100]
        else:
            return [None]

    # *** Optionally override public functions. ***



# import libraries
import json
from typing import Any, Dict, List, Optional, Type
# import local files
from ogd.core.generators.extractors.builtin.BuiltinExtractor import BuiltinExtractor
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

class CountEvent(BuiltinExtractor):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    _target_event = "NO EVENTS"
    def __init__(self, params:GeneratorParameters, schema_args:Dict[str,Any]):
        super().__init__(params=params, schema_args=schema_args)
        self._count = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _createDerivedGenerator(cls, params:GeneratorParameters, schema_args:Dict[str,Any]) -> Type[BuiltinExtractor]:
        return type(params._name, (CountEvent,), {"_target_event" : schema_args.get("target", "NO EVENTS")})

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [cls._target_event]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return []

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        if event.EventName == self._target_event:
            self._count += 1
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        ret_val : List[Any] = [self._count]
        return ret_val

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return [] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    


## import standard libraries
import abc
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.generators.Generator import Generator, GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class Extractor(Generator):
    """
    Base class for feature extractors.

    In addition to event/feature filters and an update-from-event rule,
    the Feature base class requires subclasses to implement an update-from-feature rule and a calculate rule.
    """

#TODO: use a dirty bit so we only run the GetValue function if we've received an event or feature since last calculation

    # *** ABSTRACTS ***

    ## Abstract declaration of a function to perform update of a feature from a row.
    @abc.abstractmethod
    def _updateFromFeatureData(self, feature:FeatureData):
        """Abstract declaration of a function to perform update of a feature from a row.

        :param event: An event, used to update the feature's data.
        :type event: Event
        """
        pass

    ## Abstract declaration of a function to get the calculated value of the feature, given data seen so far.
    @abc.abstractmethod
    def _getFeatureValues(self) -> List[Any]:
        """Abstract declaration of a function to get the calculated value of the feature, given data seen so far.

        :return: Returns the values of all columns for the Feature, based on data the feature has seen so far.
        :rtype: typing.Tuple
        """
        pass

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._up_to_date = False
        # by default, latest values should just be None, with length equal to number of columns for the feature.
        self._latest_values : List[Any] = [None for i in range(len(self.GetFeatureNames())) ]

    # *** PUBLIC STATICS ***

    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        """List of ExtractionMode supported by the Feature.

        Overridden from Extractor's version of the function, only makes the Feature-related modes supported.
        :return: _description_
        :rtype: List[ExtractionMode]
        """
        return [ExtractionMode.POPULATION, ExtractionMode.PLAYER, ExtractionMode.SESSION]

    @staticmethod
    def FeatureDependencyModes() -> List[ExtractionMode]:
        """List of ExtractionModes .

        Overridden from Extractor's version of the function, only makes the Feature-related modes supported.
        :return: _description_
        :rtype: List[ExtractionMode]
        """
        return [ExtractionMode.POPULATION, ExtractionMode.PLAYER, ExtractionMode.SESSION]

    # *** PUBLIC METHODS ***

    def ToFeatureData(self, player_id:Optional[str]=None, sess_id:Optional[str]=None) -> FeatureData:
        return FeatureData(
            name=self.Name,
            feature_type=type(self).__name__,
            count_index=self.CountIndex,
            cols=self.GetFeatureNames(),
            vals=self.GetFeatureValues(),
            mode=self.ExtractionMode,
            player_id=player_id,
            sess_id=sess_id
        )

    def BaseFeatureSuffix(self) -> str:
        """Base function to add a suffix to the base feature name, which will not affect the naming of subfeatures.
        By default, returns ""; override to set a suffix.
        Example use-case: Suppose you want a feature that captures the number of times a player enters a given state,
        as well as the time spent in that state.
        The feature can be named "State", with BaseFeatureSuffix returning "EntryCount" and a subfeature named "Time."
        Then the columns will be named "StateEntryCount" and "StateTime."

        :return: _description_
        :rtype: str
        """
        return ""

    def Subfeatures(self) -> List[str]:
        """Base function to get a list of names of the sub-feature(s) a given Feature class outputs.
        By default, a Feature class has no subfeatures.
        However, if a Feature class is written to output multiple values, it will need to override this function to return an appropriate list.
        Note, Subfeatures **must** match the ordering from the override of GetFeatureNames, if returning a list of length > 0.

        :return: A list of names of subfeatures for the Feature sub-class.
        :rtype: Tuple[str]
        """
        return []

    def GetFeatureNames(self) -> List[str]:
        """Base function to get a list of names of the feature(s) a given Feature class outputs.
        By default, a Feature class just generates one value, and uses its own name (defined in the schema.json file).
        If Subfeatures was overridden, and returns a non-empty list, there will be additional feature names in the list this function returns.
        Each subfeature will have the base feature's name as a prefix.

        :return: [description]
        :rtype: List[str]
        """
        return [f"{self.Name}{self.BaseFeatureSuffix()}"] + [f"{self.Name}-{subfeature}" for subfeature in self.Subfeatures()]

    def UpdateFromEvent(self, event:Event):
        """Overridden version of function from Extractor base class;
        In this case, set the "up to date" value to False whenever we see a new event.

        :param event: _description_
        :type event: Event
        """
        if self._validateEvent(event=event):
            self._updateFromEvent(event=event)
            self._up_to_date = False

    def UpdateFromFeatureData(self, feature:FeatureData):
        # TODO: add validation for FeatureData, if applicable/possible.
        # TODO: figure out a way to invalidate/reset if more events are given to features on which the given feature depends.
        self._updateFromFeatureData(feature=feature)
        self._up_to_date = False

    def GetFeatureValues(self) -> List[Any]:
        # Only call calculation feature if new events were seen since last call.
        # Practically, this doesn't matter because we always process all data before using GetFeatureValues.
        # Someday, however, this may be useful when dealing with a caching system.
        if not self._up_to_date:
            self._latest_value = self._getFeatureValues()
            self._up_to_date = True
        return self._latest_value

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***


"""Initializer for Wave Combinator features"""

__all__ = [
    "TotalSliderMoves",
    "TotalArrowMoves",
    "TotalLevelTime",
    "SliderAverageStandardDeviations",
    "TotalMoveTypeChanges",
    "SliderAverageRange",
    "TotalResets",
    "TotalFails",
    "TotalSkips",
    "MenuButtonCount",
    "BeginCount",
    "SucceedCount",
    "SessionSucceedCount",
    "Completed",
    "PercentAmplitudeMoves",
    "PercentOffsetMoves",
    "PercentWavelengthMoves",
    "AmplitudeGoodMoveCount",
    "OffsetGoodMoveCount",
    "WavelengthGoodMoveCount",
    "PercentAmplitudeGoodMoves",
    "PercentOffsetGoodMoves",
    "PercentWavelengthGoodMoves",
    "FirstMoveType",
    "ClosenessSlope",
    "ClosenessIntercept",
    "ClosenessR2",
    "RangeSlope",
    "RangeIntercept",
    "RangeR2",
    "QuestionAnswered",
    "QuestionCorrect",
    "TimeToAnswerMS",
    "SessionID",
    "PersistentSessionID",
    "AverageSliderMoves",
    "AverageLevelTime",
    "AverageMoveTypeChanges",
    "OverallSliderAverageStandardDeviations",
    "OverallSliderAverageRange",
    "OverallPercentAmplitudeMoves",
    "OverallPercentOffsetMoves",
    "OverallPercentWavelengthMoves",
    "AverageFails",
    "SequenceLevel"
]

from . import TotalSliderMoves
from . import TotalArrowMoves
from . import TotalLevelTime
from . import SliderAverageStandardDeviations
from . import TotalMoveTypeChanges
from . import SliderAverageRange
from . import TotalResets
from . import TotalFails
from . import TotalSkips
from . import MenuButtonCount
from . import BeginCount
from . import SucceedCount
from . import SessionSucceedCount
from . import Completed
from . import PercentAmplitudeMoves
from . import PercentOffsetMoves
from . import PercentWavelengthMoves
from . import AmplitudeGoodMoveCount
from . import OffsetGoodMoveCount
from . import WavelengthGoodMoveCount
from . import PercentAmplitudeGoodMoves
from . import PercentOffsetGoodMoves
from . import PercentWavelengthGoodMoves
from . import FirstMoveType
from . import ClosenessSlope
from . import ClosenessIntercept
from . import ClosenessR2
from . import RangeSlope
from . import RangeIntercept
from . import RangeR2
from . import QuestionAnswered
from . import QuestionCorrect
from . import TimeToAnswerMS
from . import SessionID
from . import PersistentSessionID
from . import AverageSliderMoves
from . import AverageLevelTime
from . import AverageMoveTypeChanges
from . import OverallSliderAverageStandardDeviations
from . import OverallSliderAverageRange
from . import OverallPercentAmplitudeMoves
from . import OverallPercentOffsetMoves
from . import OverallPercentWavelengthMoves
from . import AverageFails
from . import SequenceLevel



import logging
import json
from datetime import datetime
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class SnippetsSubmitted(PerLevelFeature):
    has_printed = False
    def __init__(self, params:GeneratorParameters):
        PerLevelFeature.__init__(self, params=params)
        self._snippet_ids : List[str] = []


    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["story_click"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "story_click":
            snippet_list = json.loads( event.EventData["snippet_list"] )
            for snippet in snippet_list:
                # if not SnippetsSubmitted.has_printed:
                #     print(f"snippet: {snippet} of type {type(snippet)}")
                #     SnippetsSubmitted.has_printed = True
                self._snippet_ids.append(snippet.get("SnippetId", "NOT FOUND"))

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._snippet_ids]

    # *** Optionally override public functions. ***
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.SESSION, ExtractionMode.PLAYER]




from typing import Any, List

from typing import Any, Dict, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class Session_Language(SessionFeature):
    """_summary_

    :param SessionFeature: _description_
    :type SessionFeature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        self._session_language = None
        super().__init__(params=params)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ['language_selected']

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        self._session_language = event.EventData.get("language")

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._session_language]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "2"


# import libraries
import logging
from datetime import datetime
from ogd.core.models import Event
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class AverageLevelTime(SessionFeature):
    def __init__(self, params:GeneratorParameters):
        SessionFeature.__init__(self, params=params)
        self._levels_encountered : set                      = set()
        self._begin_times        : Dict[int,List[datetime]] = {}
        self._complete_times     : Dict[int,List[datetime]] = {}

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["BEGIN.0", "COMPLETE.0"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        _level = event.GameState['level']
        self._levels_encountered.add(_level) # set-add level to list, at end we will have set of all levels seen.
        if event.EventName == "BEGIN.0":
            if not _level in self._begin_times.keys():
                self._begin_times[_level] = []
            self._begin_times[_level].append(event.Timestamp)
        elif event.EventName == "COMPLETE.0":
            if not _level in self._complete_times.keys():
                self._complete_times[_level] = []
            self._complete_times[_level].append(event.Timestamp)
        else:
            Logger.Log(f"AverageLevelTime received an event which was not a BEGIN or a COMPLETE!", logging.WARN)

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if len(self._begin_times) < len(self._complete_times):
            self.WarningMessage(f"Player began level {self.CountIndex} {len(self._begin_times)} times but completed it {len(self._complete_times)}.")
        _diffs      = []
        for level in self._levels_encountered:
            if level in self._begin_times.keys() and level in self._complete_times.keys():
                _num_plays  = min(len(self._begin_times[level]), len(self._complete_times[level]))
                _diffs += [(self._complete_times[level][i] - self._begin_times[level][i]).total_seconds() for i in range(_num_plays)]
            else:
                if level not in self._begin_times.keys() and level in self._complete_times.keys():
                    Logger.Log(f"Player completed level {level}, but did not begin it!", logging.WARN)
                elif level in self._begin_times.keys() and level not in self._complete_times.keys():
                    Logger.Log(f"Player began level {level}, but did not complete it.", logging.DEBUG)
                elif level not in self._begin_times.keys() and level not in self._complete_times.keys():
                    Logger.Log(f"Player had level {level} listed as encountered, but did not begin *or* complete it.", logging.WARN)
        _total_time = sum(_diffs)
        if len(self._levels_encountered) > 0:
            return [_total_time / len(self._levels_encountered)]
        else:
            return [None]

    # *** Optionally override public functions. ***



from typing import List, Optional, Dict, Any
from models.SequenceModel import SequenceModel


## @class MapSummaryModel
# Returns an array (list of lists) with values indicating the tile type for the corresponding tile in the game.
# To interpret the tile type values, see https://github.com/fielddaylab/lakeland/blob/master/README.md#TileType
# If there is at least one building present (home, farm, or dairy), the map returned is just a bounding box
# with a border width of five tiles around the area the buildings occupy.
# Otherwise, the map returned is the entire map.
# @param levels: Levels applicable for model
class MapSummaryModel(SequenceModel):
    def __init__(self, levels: List[int] = []):
        '''
        @class MapSummaryModel
        Returns an array (list of lists) with values indicating the tile type for the corresponding tile in the game.
        To interpret the tile type values, see https://github.com/fielddaylab/lakeland/blob/master/README.md#TileType
        If there is at least one building present (home, farm, or dairy), the map returned is just a bounding box
        with a border width of five tiles around the area the buildings occupy.
        Otherwise, the map returned is the entire map.
        :param levels: Levels applicable for model
        '''
        super().__init__()

    def _eval(self, events: List[Dict[str, Any]], verbose: bool = False) -> List[List[int]]:
        assert events
        # compute the full bitmap
        w, h = 50, 50;
        bitmap = [[-1 for x in range(w)] for y in range(h)]
        for event in reversed(events):
            if event["event_custom"] == 0:
                tile_string = event["event_data_complex"]["tiles"]
                _tile = self._read_stringified_array(tile_string)
                tile = self._array_to_mat(4, _tile)
                for i in range(0, len(tile)):
                    bitmap[int(i / 50)][i % 50] = tile[i][3]
                break
        # shrink the map down to only a border of five tiles around the buildings
        i_coords = []
        j_coords = []
        for i in range(50):
            for j in range(50):
                if bitmap[i][j] in [8, 9, 10]:
                    i_coords.append(i)
                    j_coords.append(j)
        if len(i_coords) != 0:
            # compute the min and max i,j coordinates, but if they are closer than five tiles
            # to the edge of the map, set them to exactly five tiles from the edge
            max_i = min([max(i_coords), 44])
            min_i = max([min(i_coords), 5])
            max_j = min([max(j_coords), 44])
            min_j = max([min(j_coords), 5])
            small_bitmap = [[bitmap[i][j] for i in range(min_i - 5, max_i + 5)] for j in range(min_j - 5, max_j + 5)]
        else:
            small_bitmap = bitmap

        return small_bitmap

    # reformat raw variable functions
    def _read_stringified_array(self, arr: str) -> List[int]:
        if not arr:
            return []
        return [int(x) for x in arr.split(',')]


    def _array_to_mat(self, num_columns: int, arr: List[int]) -> List[List[int]]:
        assert len(arr) % num_columns == 0
        return [arr[i:i + num_columns] for i in range(0, len(arr), num_columns)]


def __repr__(self):
        return f"MapSummaryModel(levels={self._levels}, input_type={self._input_type})"


# import standard libraries
import itertools
import json
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Dict, Final, List, Optional
# import OGD files
from ogd.games import AQUALAB
from ogd.games.AQUALAB.detectors import *
from ogd.games.AQUALAB.features import *
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils import utils
from ogd.core.utils.Logger import Logger
from . import features

EXPORT_PATH : Final[str] = "games/AQUALAB/DBExport.json"

class AqualabLoader(GeneratorLoader):
    """Loader subclass for Aqualab generators"""

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the AqualabLoader class.
    def __init__(self, player_id:str, session_id:str, game_schema: GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        """Constructor for the AqualabLoader class.

        :param player_id: The player ID associated with the feature extractors/event detectors being loaded.
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)
        self._job_map = {"no-active-job": 0}
        self._diff_map = {0: {"experimentation": 0, "modeling": 0, "argumentation": 0} }
        self._task_map = {}

        # Load Aqualab jobs export and map job names to integer values
        _dbexport_path = Path(AQUALAB.__file__) if Path(AQUALAB.__file__).is_dir() else Path(AQUALAB.__file__).parent
        with open(_dbexport_path / "DBExport.json", "r") as file:
            export = json.load(file)

            task_num = 1
            for i, job in enumerate(export["jobs"], start=1):
                self._job_map[job["id"]] = i
                self._diff_map[i] = job["difficulties"]
                for task in job["tasks"]:
                    task_by_job = job["id"] + "_" + task["id"]
                    self._task_map[task_by_job] = task_num
                    task_num += 1

        # Update level count
        self._game_schema._max_level = len(self._job_map) - 1

    @property
    def JobMap(self) -> Dict:
        """Property returning the "job map" that assigns each job in the DBExport an integer value/index.

        :return: The "job map" from the current DBExport
        :rtype: Dict
        """
        return self._job_map

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule() -> ModuleType:
        return features

    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Optional[Feature]:
        ret_val : Optional[Feature] = None
        # First run through aggregate features
        if extractor_params._count_index is None:
            match feature_type:
                case "ActiveTime":
                    ret_val = ActiveTime.ActiveTime(params=extractor_params, job_map=self._job_map, active_threads=schema_args.get("Active_threshold"))
                case "ActiveJobs":
                    ret_val = ActiveJobs.ActiveJobs(params=extractor_params, job_map=self._job_map)
                case "AppVersions":
                    ret_val = AppVersions.AppVersions(params=extractor_params)
                case "AverageSessionTime":
                    ret_val = AverageSessionTime.AverageSessionTime(params=extractor_params)
                case "EchoSessionID":
                    ret_val = EchoSessionID.EchoSessionID(params=extractor_params)
                case "EventList":
                    ret_val = EventList.EventList(params=extractor_params)
                case "JobsCompleted":
                    ret_val = JobsCompleted.JobsCompleted(params=extractor_params, player_id=self._player_id)
                case "JobTriesInArgument":
                    ret_val = JobTriesInArgument.JobTriesInArgument(params=extractor_params, job_map=self._job_map)
                case "JobPlayTime":
                    ret_val = JobPlayTime.JobPlayTime(params=extractor_params, job_map=self._job_map)
                case "ModelExportCount":
                    ret_val = ModelExportCount.ModelExportCount(params=extractor_params, job_map=self._job_map)
                case "ModelInterveneCount":
                    ret_val = ModelInterveneCount.ModelInterveneCount(params=extractor_params, job_map=self._job_map)
                case "ModelPredictCount":
                    ret_val = ModelPredictCount.ModelPredictCount(params=extractor_params, job_map=self._job_map)
                # case "PlayLocations":
                #     ret_val = PlayLocations.PlayLocations(params=extractor_params)
                case "PlayerSummary":
                    ret_val = PlayerSummary.PlayerSummary(params=extractor_params)
                case "PopulationSummary":
                    ret_val = PopulationSummary.PopulationSummary(params=extractor_params)
                case "SessionDiveSitesCount":
                    ret_val = SessionDiveSitesCount.SessionDiveSitesCount(params=extractor_params)
                case "SessionDuration":
                    ret_val = SessionDuration.SessionDuration(params=extractor_params, threshold=int(schema_args.get("threshold", 60)))
                case "SessionGuideCount":
                    ret_val = SessionGuideCount.SessionGuideCount(params=extractor_params)
                case "SessionHelpCount":
                    ret_val = SessionHelpCount.SessionHelpCount(params=extractor_params)
                case "SessionID":
                    ret_val = SessionID.SessionID(params=extractor_params, session_id=self._session_id)
                case "SessionJobsCompleted":
                    ret_val = SessionJobsCompleted.SessionJobsCompleted(params=extractor_params)
                case "SwitchJobsCount":
                    ret_val = SwitchJobsCount.SwitchJobsCount(params=extractor_params)
                case "TankRulesCount":
                    ret_val = TankRulesCount.TankRulesCount(params=extractor_params)
                case "TopJobCompletionDestinations":
                    ret_val = TopJobCompletionDestinations.TopJobCompletionDestinations(params=extractor_params, job_map=self._job_map)
                case "TopJobSwitchDestinations":
                    ret_val = TopJobSwitchDestinations.TopJobSwitchDestinations(params=extractor_params, job_map=self._job_map)
                case "TotalArcticTime":
                    ret_val = TotalArcticTime.TotalArcticTime(params=extractor_params)
                case "TotalArgumentationTime":
                    ret_val = TotalArgumentationTime.TotalArgumentationTime(params=extractor_params)
                case "TotalBayouTime":
                    ret_val = TotalBayouTime.TotalBayouTime(params=extractor_params)
                case "TotalCoralTime":
                    ret_val = TotalCoralTime.TotalCoralTime(params=extractor_params)
                case "TotalDiveTime":
                    ret_val = TotalDiveTime.TotalDiveTime(params=extractor_params)
                case "TotalExperimentationTime":
                    ret_val = TotalExperimentationTime.TotalExperimentationTime(params=extractor_params)
                case "TotalGuideCount":
                    ret_val = TotalGuideCount.TotalGuideCount(params=extractor_params)
                case "TotalHelpCount":
                    ret_val = TotalHelpCount.TotalHelpCount(params=extractor_params)
                case "TotalKelpTime":
                    ret_val = TotalKelpTime.TotalKelpTime(params=extractor_params)
                case "TotalModelingTime":
                    ret_val = TotalModelingTime.TotalModelingTime(params=extractor_params)
                case "TotalPlayTime":
                    ret_val = TotalPlayTime.TotalPlayTime(params=extractor_params)
                case "UserAvgActiveTime":
                    ret_val = UserAvgActiveTime.UserAvgActiveTime(params=extractor_params, player_id=self._player_id)
                case "UserAvgSessionDuration":
                    ret_val = UserAvgSessionDuration.UserAvgSessionDuration(params=extractor_params, player_id=self._player_id)
                case "UserTotalSessionDuration":
                    ret_val = UserTotalSessionDuration.UserTotalSessionDuration(params=extractor_params, player_id=self._player_id)
                case _:
                    Logger.Log(f"'{feature_type}' is not a valid aggregate feature type for Aqualab.")
        # then run through per-count features.
        else:
            match feature_type:
                case "JobActiveTime":
                    ret_val = JobActiveTime.JobActiveTime(params=extractor_params, job_map=self._job_map)
                case "JobArgumentation":
                    ret_val = JobArgumentation.JobArgumentation(params=extractor_params, job_map=self._job_map)
                case "JobCompletionTime":
                    ret_val = JobCompletionTime.JobCompletionTime(params=extractor_params, job_map=self._job_map)
                case "JobDiveSitesCount":
                    ret_val = JobDiveSitesCount.JobDiveSitesCount(params=extractor_params, job_map=self._job_map)
                case "JobDiveTime":
                    ret_val = JobDiveTime.JobDiveTime(params=extractor_params, job_map=self._job_map)
                case "JobExperimentation":
                    ret_val = JobExperimentation.JobExperimentation(params=extractor_params, job_map=self._job_map)
                case "JobGuideCount":
                    ret_val = JobGuideCount.JobGuideCount(params=extractor_params, job_map=self._job_map)
                case "JobHelpCount":
                    ret_val = JobHelpCount.JobHelpCount(params=extractor_params, job_map=self._job_map)
                case "JobLocationChanges":
                    ret_val = JobLocationChanges.JobLocationChanges(params=extractor_params, job_map=self._job_map)
                case "JobModeling":
                    ret_val = JobModeling.JobModeling(params=extractor_params, job_map=self._job_map)
                case "JobPriorAttempt":
                    ret_val = JobPriorAttempt.JobPriorAttempt(params=extractor_params, job_map=self._job_map)
                case "JobPriorComplete":
                    ret_val = JobPriorComplete.JobPriorComplete(params=extractor_params, job_map=self._job_map)
                case "JobStartCount":
                    ret_val = JobStartCount.JobStartCount(params=extractor_params, job_map=self._job_map)
                case "JobTasksCompleted":
                    ret_val = JobTasksCompleted.JobTasksCompleted(params=extractor_params, job_map=self._job_map)
                case "JobsAttempted":
                    ret_val = JobsAttempted.JobsAttempted(params=extractor_params, job_map=self._job_map, diff_map=self._diff_map)
                case "RegionName":
                    ret_val = RegionName.RegionName(params=extractor_params)
                case "RegionJobCount":
                    ret_val = RegionJobCount.RegionJobCount(params=extractor_params)
                case "SyncCompletionTime":
                    ret_val = SyncCompletionTime.SyncCompletionTime(params=extractor_params)
                case _:
                    Logger.Log(f"'{feature_type}' is not a valid per-count feature type for Aqualab.")
        return ret_val

    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Optional[Detector]:
        ret_val : Optional[Detector] = None

        match detector_type:
            case "CollectFactNoJob":
                ret_val = CollectFactNoJob.CollectFactNoJob(params=extractor_params, trigger_callback=trigger_callback)
            case "DiveSiteNoEvidence":
                ret_val = DiveSiteNoEvidence.DiveSiteNoEvidence(params=extractor_params, trigger_callback=trigger_callback, threshold=schema_args['threshold'])
            case "EchoRoomChange":
                ret_val = EchoRoomChange.EchoRoomChange(params=extractor_params, trigger_callback=trigger_callback)
            case "Idle":
                ret_val = Idle.Idle(params=extractor_params, trigger_callback=trigger_callback, idle_level=schema_args.get("idle_level"))
            case "SceneChangeFrequently":
                ret_val = SceneChangeFrequently.SceneChangeFrequently(params=extractor_params, trigger_callback=trigger_callback, time_threshold=schema_args.get("threshold"))
            case "HintAndLeave":
                ret_val = HintAndLeave.HintAndLeave(params=extractor_params, trigger_callback=trigger_callback, time_threshold=schema_args.get("threshold"))
            case "TwoHints":
                ret_val = TwoHints.TwoHints(params=extractor_params, trigger_callback=trigger_callback, time_threshold=schema_args.get("threshold"))
            case _:
                Logger.Log(f"'{detector_type}' is not a valid detector for Aqualab.")
        return ret_val

# *** PUBLIC STATICS ***

    @staticmethod
    def GetAqualabJobCount(db_export_path:Path=Path(".") / "ogd" / "games" / "AQUALAB") -> int:
        """Function to retrieve the number of jobs in the DBExport file.

        :param db_export_path: A path to the folder containing DBExport, defaults to Path(".")/"ogd"/"games"/"AQUALAB"
        :type db_export_path: Path, optional
        :return: The number of jobs in the current DBExport
        :rtype: int
        """
        ret_val : int
        db_export = utils.loadJSONFile(filename="DBExport.json", path=db_export_path)
        ret_val = len(db_export.get("jobs", []))
        return ret_val

    @staticmethod
    def GetAqualabTaskCount(db_export_path:Path=Path(".") / "ogd" / "games" / "AQUALAB") -> int:
        """Function to retrieve the total number of tasks in the DBExport file.

        :param db_export_path: A path to the folder containing DBExport, defaults to Path(".")/"ogd"/"games"/"AQUALAB"
        :type db_export_path: Path, optional
        :return: The total number of tasks in the current DBExport
        :rtype: int
        """
        db_export = utils.loadJSONFile(filename="DBExport.json", path=db_export_path)
        list_o_lists = [job.get('tasks', []) for job in db_export.get('jobs', [])]
        all_tasks    = list(itertools.chain.from_iterable(list_o_lists))
        return len(all_tasks)



"""Initializer for Aqualab features"""

__all__ = [
    "ActiveJobs",
    "ActiveTime",
    "AppVersions",
    "AverageSessionTime",
    "EchoSessionID",
    "EventList",
    "JobActiveTime",
    "JobArgumentation",
    "JobCompletionTime",
    "JobDiveSitesCount",
    "JobDiveTime",
    "JobExperimentation",
    "JobGuideCount",
    "JobHelpCount",
    "JobLocationChanges",
    "JobModeling",
    "JobPlayTime",
    "JobPriorAttempt",
    "JobPriorComplete",
    "JobsAttempted",
    "JobsCompleted",
    "JobStartCount",
    "JobTasksCompleted",
    "JobTriesInArgument",
    "ModelExportCount",
    "ModelInterveneCount",
    "ModelPredictCount",
    "PlayerSummary",
    "PopulationSummary",
    "RegionJobCount",
    "RegionName",
    "SessionDiveSitesCount",
    "SessionDuration",
    "SessionGuideCount",
    "SessionHelpCount",
    "SessionID",
    "SessionJobsCompleted",
    "SwitchJobsCount",
    "SyncCompletionTime",
    "TankRulesCount",
    "TopJobCompletionDestinations",
    "TopJobSwitchDestinations",
    "TotalArcticTime",
    "TotalArgumentationTime",
    "TotalBayouTime",
    "TotalCoralTime",
    "TotalDiveTime",
    "TotalExperimentationTime",
    "TotalGuideCount",
    "TotalHelpCount",
    "TotalKelpTime",
    "TotalModelingTime",
    "TotalPlayTime",
    "UserAvgActiveTime",
    "UserAvgSessionDuration",
    "UserTotalSessionDuration",
]

from . import ActiveJobs
from . import ActiveTime
from . import AppVersions
from . import AverageSessionTime
from . import EchoSessionID
from . import EventList
from . import JobActiveTime
from . import JobArgumentation
from . import JobCompletionTime
from . import JobDiveSitesCount
from . import JobDiveTime
from . import JobExperimentation
from . import JobGuideCount
from . import JobHelpCount
from . import JobLocationChanges
from . import JobModeling
from . import JobPlayTime
from . import JobPriorAttempt
from . import JobPriorComplete
from . import JobsAttempted
from . import JobsCompleted
from . import JobStartCount
from . import JobTasksCompleted
from . import JobTriesInArgument
from . import ModelExportCount
from . import ModelInterveneCount
from . import ModelPredictCount
from . import PlayerSummary
from . import PopulationSummary
from . import RegionJobCount
from . import RegionName
from . import SessionDiveSitesCount
from . import SessionDuration
from . import SessionGuideCount
from . import SessionHelpCount
from . import SessionID
from . import SessionJobsCompleted
from . import SwitchJobsCount
from . import SyncCompletionTime
from . import TankRulesCount
from . import TopJobCompletionDestinations
from . import TopJobSwitchDestinations
from . import TotalArcticTime
from . import TotalArgumentationTime
from . import TotalBayouTime
from . import TotalCoralTime
from . import TotalDiveTime
from . import TotalExperimentationTime
from . import TotalGuideCount
from . import TotalHelpCount
from . import TotalKelpTime
from . import TotalModelingTime
from . import TotalPlayTime
from . import UserAvgActiveTime
from . import UserAvgSessionDuration
from . import UserTotalSessionDuration



## import standard libraries
from datetime import datetime
import re
from typing import Any, Callable, Dict, List, Optional
## import local files
from . import features
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.games.JOWILDER.features import *
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.Logger import Logger

class JowilderLoader(GeneratorLoader):

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule():
        return features

    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Optional[Feature]:
        ret_val : Optional[Feature] = None
        match feature_type:
            case "QuestionAnswers":
                ret_val = QuestionAnswers.QuestionAnswers(params=extractor_params)
            case "SurveyItem":
                ret_val = SurveyItem.SurveyItem(params=extractor_params)
            case "Interaction":
                ret_val = Interaction.Interaction(params=extractor_params)
            case "SurveyTime":
                ret_val = SurveyTime.SurveyTime(params=extractor_params)
            case "Clicks":
                ret_val = Clicks.Clicks(params=extractor_params)
            case "Hovers":
                ret_val = Hovers.Hovers(params=extractor_params)
            case "SessionDuration":
                ret_val = SessionDuration.SessionDuration(params=extractor_params)
            case "InteractionName":
                ret_val = InteractionName.InteractionName(params=extractor_params)
            case "NotebookUses":
                ret_val = NotebookUses.NotebookUses(params=extractor_params)
            case "EventCount":
                ret_val = EventCount.EventCount(params=extractor_params)
            case "UserEnabled":
                ret_val = UserEnabled.UserEnabled(params=extractor_params)
            case "GameVersion":
                ret_val = GameVersion.GameVersion(params=extractor_params)
            case "UsedSaveCode":
                ret_val = UsedSaveCode.UsedSaveCode(params=extractor_params)
            case "GameScript":
                ret_val = GameScript.GameScript(params=extractor_params)
            case "SessionStart":
                ret_val = SessionStart.SessionStart(params=extractor_params)
            case "IdleState":
                ret_val = IdleState.IdleState(params=extractor_params, threshold=schema_args.get("IDLE_THRESH_SECONDS", IdleState.IdleState.IDLE_TIME_THRESHOLD))
            case "ActiveStateTime":
                ret_val = ActiveStateTime.ActiveStateTime(params=extractor_params, threshold=schema_args.get('ACTIVE_THRESH_SECONDS', ActiveStateTime.ActiveStateTime.ACTIVE_TIME_THRESHOLD))
            case "MeaningfulActions":
                ret_val = MeaningfulActions.MeaningfulActions(params=extractor_params)
            case "FirstInteraction":
                ret_val = FirstInteraction.FirstInteraction(params=extractor_params)
            case "LastInteraction":
                ret_val = LastInteraction.LastInteraction(params=extractor_params)
            case "UsedContinue":
                ret_val = UsedContinue.UsedContinue(params=extractor_params)
            case "InteractionWordsPerSecond":
                ret_val = InteractionWordsPerSecond.InteractionWordsPerSecond(params=extractor_params)
            case "InteractionTextBoxesPerSecond":
                ret_val = InteractionTextBoxesPerSecond.InteractionTextBoxesPerSecond(params=extractor_params)
            case _:
                Logger.Log(f"'{feature_type}' is not a valid feature for Jowilder.")
        return ret_val

    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Optional[Detector]:
        Logger.Log(f"'{detector_type}' is not a valid feature for Lakeland.")
        return None

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the JoWilderExtractor class.
    def __init__(self, player_id:str, session_id:str, game_schema: GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        """Constructor for the CrystalLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)



## import standard libraries
import abc
from typing import Dict, List, Type, Optional, Set

# import locals
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.registries.ExtractorRegistry import ExtractorRegistry
from ogd.core.processors.GeneratorProcessor import GeneratorProcessor
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.utils.utils import ExportRow

## @class Processor
class ExtractorProcessor(GeneratorProcessor):

    # *** ABSTRACTS ***

    ## Abstract declaration of a function to get the calculated value of the feature, as a FeatureData package, given data seen so far.
    @abc.abstractmethod
    def _getFeatureData(self, order:int) -> List[FeatureData]:
        pass

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_schema: GameSchema, LoaderClass:Type[GeneratorLoader], feature_overrides:Optional[List[str]]=None):
        super().__init__(game_schema=game_schema, LoaderClass=LoaderClass, feature_overrides=feature_overrides)
        self._registry : ExtractorRegistry = ExtractorRegistry(mode=self._mode)
        self._registry.LoadFromSchema(schema=game_schema, loader=self._loader, overrides=feature_overrides)

    def __str__(self):
        return f""

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def GetFeatureData(self, order:int) -> List[FeatureData]:
        # TODO: add error handling code, if applicable.
        return self._getFeatureData(order=order)

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
import json
from typing import Any, List, Optional
# import local files
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.games.ICECUBE.DBExport import scene_map
from ogd.games.ICECUBE.features.PerSceneFeature import PerSceneFeature
from ogd.games.ICECUBE.DBExport import scenes_list
from ogd.core.utils.Logger import Logger
from datetime import timedelta

# import libraries
import logging

scenes_map = {"ICE":0, "VOYAGER":1, "NOTHING":2, "EXTREME":3, "EARTH":4, "CREDITS":5, "None":6}

class SceneDuration(PerCountFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._session_id = None
        self._scene_start_time = None
        self._prev_timestamp = None
        self._time = 0
        self._name = None
    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["scene_begin","scene_end"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.SessionID != self._session_id:
            self._session_id = event.SessionID

            if self._scene_start_time and self._prev_timestamp:
                self._time += (self._prev_timestamp - self._scene_start_time).total_seconds()
                self._scene_start_time = event.Timestamp

        
        if event.EventName == "scene_begin":
            self._scene_start_time = event.Timestamp
        elif event.EventName == "scene_end":
            if self._scene_start_time is not None:
                self._time += (event.Timestamp - self._scene_start_time).total_seconds()
                self._scene_start_time = None

        self._prev_timestamp = event.Timestamp

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [timedelta(seconds=self._time)]

    # *** Optionally override public functions. ***
    def _validateEventCountIndex(self, event: Event):
        ret_val : bool = False

        scene_data = event.game_state.get("scene_name")
        # if scene_data in scenes_map and scenes_map[scene_data] == self.CountIndex:
        #         ret_val = True
        if scene_data is not None:
            if scene_data in scenes_map and scenes_map[scene_data] == self.CountIndex:
                ret_val = True
        else:
            self.WarningMessage(f"Got invalid scene data in {type(self).__name__}")

        return ret_val
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "2"


# import libraries
import json
import logging
from datetime import datetime, timedelta
from typing import Any, Dict, Final, List, Optional
# import local files
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event
from ogd.core.utils.Logger import Logger

# BUG: Question0 and quiz 0 don't have start time
# NOTE: Assumptions are: Every quiz should have a quizstart.

QUIZ_INDEXES : Final[Dict[int, Dict[int, int]]] = {
    0: {0: 0, 1: 1},
    2: {0: 2, 1: 3, 2: 4, 3: 5},
    3: {0: 6, 1: 7, 2: 8, 3: 9},
    4: {0: 10, 1: 11, 2: 12, 3: 13},
    5: {0: 14, 1: 15, 2: 16, 3: 17}
}

class SurveyItem(PerCountFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    
    
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        # NOTE: The time we record it the first time that players answered the given question minus the last time answering the last question or the time starting the quiz
        self._last_timestamp : Optional[datetime] = None
        self._quiz_start_timestamp : Optional[datetime] = None
        self._time : Optional[timedelta] = None
        self._response_index : Optional[int] = None
        self._index : Optional[int] = None
        self._text : str = ""
        self._num_answers : int = 0
        

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    def _validateEventCountIndex(self, event: Event):
        # BUG: It's so weird that question NO.0, 4, 8... use quizstart time but others use last question's time. see JowilderExtractor.py line 492
        # TODO: In the old code, when the quiz starts, self._last_timestamp is assigned a value, and when the quiz ends, the member is assigned to None. But logically, I don't think the latter one is needed, since after a quiz ends and before another starts, we shouldn't have any other quizquestion events or access this feature extractor. So I ignored this part.
        quiz_index = event.event_data["quiz_number"]
        if event.EventName == "CUSTOM.23":
            return self.CountIndex in QUIZ_INDEXES[quiz_index].values()
        question_index = event.event_data["question_index"]
        self._index = QUIZ_INDEXES[quiz_index][question_index]
        return self._index == self.CountIndex or self._index == self.CountIndex - 1

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["CUSTOM.22", "CUSTOM.23"] 
        # ["CUSTOM.22", "CUSTOM.23"] = ["quizquestion", "quizstart"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] 

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        # TODO: Add explict code and comments to skip quiz 0 also for survey_time
        if self.CountIndex == 0:
            return 
        if event.EventName == "CUSTOM.23":
            self._quiz_start_timestamp = event.Timestamp
            return
        if self._index == self.CountIndex - 1: 
            self._last_timestamp = event.Timestamp
            return
        
        if self._time in [0, None]:
            _question_index = event.EventData.get("question_index")
            if _question_index == 0:
                if self._quiz_start_timestamp is not None:
                    self._time = event.Timestamp - self._quiz_start_timestamp
                else:
                    raise ValueError(f"SurveyItem got a question answer when there was no quiz start time.")
            else:
                if self._last_timestamp is not None:
                    self._time = event.Timestamp - self._last_timestamp
                else:
                    self._time = None
                    Logger.Log("SurveyItem got a question answer when there was no last timestamp.", logging.DEBUG)
                    # raise ValueError(f"SurveyItem got a question answer when there was no last timestamp.")
        self._response_index = event.EventData["response_index"]
        self._text = event.EventData["response"]
        self._num_answers += 1
            
        
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        # By now, ResponseChanges will return -1 if there is no answer for this question.
        return [self._response_index, self._text, self._time, self._num_answers - 1]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Text", "Time", "ResponseChanges"]



# import libraries
import json
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional
# import local files
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event
from ogd.games.JOWILDER import Jowilder_Enumerators as je
from ogd.core.utils.utils import loadJSONFile
from ogd.core.utils.Logger import Logger

interaction_path = Path('.') / "ogd" / "games" / "JOWILDER"
METADATA_RAW : Dict[str, Dict[str, Any]] = loadJSONFile(filename="interaction_metadata.json", path=interaction_path, search_in_src=True)
METADATA     : Dict[int, Dict[str, Any]] = {je.fqid_to_enum.get(v.get("fqid", "FQID NOT FOUND"), -1): v for v in METADATA_RAW.values()}
Logger.Log(f"JoWilder InteractionName loaded {len(METADATA.keys())} metadata items.")

class InteractionName(PerCountFeature):

    def __init__(self, params=GeneratorParameters):
        super().__init__(params=params)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    def _validateEventCountIndex(self, event: Event):
        return False

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return [] 

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        cur_interaction : Optional[Dict[str, Any]] = METADATA.get(self.CountIndex)
        if cur_interaction is not None:
            ret_val : List[Any] = [cur_interaction.get("fqid"), cur_interaction.get("count_boxes"), cur_interaction.get("num_words")]
        else:
            raise ValueError(f"Could not find metadata for interaction #{self.CountIndex}")
        return ret_val

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["BoxesCount", "WordsCount"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    



# import libraries
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class AverageMoveTypeChanges(SessionFeature):
    def __init__(self, params:GeneratorParameters):
        SessionFeature.__init__(self, params=params)
        self._levels_encountered : set = set()
        self._last_move = {}
        self._change_count = {}

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.1", "CUSTOM.2"]
        # return ["SLIDER_MOVE_RELEASE", "ARROW_MOVE_RELEASE"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        _level = event.GameState['level']
        self._levels_encountered.add(_level) # set-add level to list, at end we will have set of all levels seen.
        if not _level in self._change_count.keys():
            self._change_count[_level] = 0
            self._last_move[_level] = None
        if self._last_move[_level] != event.EventName:
            self._change_count[_level] += 1
        self._last_move[_level] = event.EventName

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        _counts = [count for count in self._change_count.values()]
        if len(self._levels_encountered) > 0:
            return [sum(_counts) / len(self._levels_encountered)]
        else:
            return [None]

    # *** Optionally override public functions. ***





from typing import List, Optional, Dict, Any
import pandas as pd
from datetime import datetime as dt
import json
from models.SequenceModel import SequenceModel

GAMESTATE_EVT = 0
BUY_EVT = 7
FARMFAIL_EVT = 25
FARMGROWTH_EVT = 41
BUY_FERT_EVT = 4
BUY_FARM_EVT = 3
TILE_TYPE_FARM = 9

log_type_dict = {
    GAMESTATE_EVT: "gamestate",
    BUY_EVT: "fertilizer",
    FARMFAIL_EVT: "farmfail",
    FARMGROWTH_EVT: "farmgrowth"
}

## @class SimpleFarmAbandonmentModel
# Returns a dictionary of farms abandoned and amount of time it is abandoned
# @param levels: Levels applicable for model

class SimpleFarmAbandonmentModel(SequenceModel):
    def __init__(self, levels: List[int] = []):
        super().__init__()

    def _eval(self, events: List[Dict[str, Any]], verbose: bool = False):
        self.df = pd.DataFrame.from_records(events)
        df_filter = ['app_id', 'app_id_fast', 'player_id', 'app_version',
                     'persistent_sess_id', 'event', 'event_data_simple',
                     'server_time', 'remote_addr', 'req_id', 'http_user_agent']
        if set(df_filter).issubset(set(self.df.columns)):
            self.df = self.df.drop(columns=df_filter, axis=0)
        self.df['utc_time_secs'] = pd.Series(self.df['client_time'].apply(
            lambda x: dt.fromisoformat(x).timestamp()), dtype=int)
        farm_fails = self.process_feature(FARMFAIL_EVT)
        return self.compose_results(self.add_time_abd(self.process_feature(BUY_EVT, False, BUY_FARM_EVT), farm_fails))

    def buy_filter(self, filt_type, args):
        res_filt = []
        if filt_type == BUY_EVT:
            for _ in self.df[self.df['event_custom'] == BUY_EVT].itertuples():
                if (_.event_data_complex['buy'] != args[1]) or not _.event_data_complex['success']:
                    res_filt.append(_.Index)
        return res_filt

    def feature_extract(self, evt):
        return self.df[self.df['event_custom'] == evt]

    def join_complex(self, frame):
        res_df = frame.join(
            pd.DataFrame(frame.event_data_complex.tolist(), index=frame.index),
            rsuffix='_complex')
        if 'utc_time_secs_complex' in list(res_df.columns):
            res_df = res_df.drop('utc_time_secs_complex', axis=1)
            'event_custom_complex'
        if 'event_custom_complex' in list(res_df.columns):
            res_df = res_df.drop('event_custom_complex', axis=1)
        return res_df

    def dict_tile(self, tile, farm_id):
        return {
            "val": tile[0],
            "nutrition": tile[1],
            "og_type": tile[2],
            "curr_type": tile[3],
            "tx": tile[4],
            "ty": tile[5],
            "farm_id": farm_id
        }

    def add_tiles(self, frame, is_gamestate=False):
        if len(frame.index) == 0:
            return frame

        if is_gamestate:
            # frame['tiles'] = frame['tiles'].apply(lambda x: compile_gs_tiles(x))
            pass
        else:
            frame['tile'] = frame['tile'].apply(lambda x: self.dict_tile(x, 0))
        return frame

    def process_feature(self, evt, is_gamestate=False, buy_item=0):
        if evt == BUY_EVT:
            return self.add_tiles(self.join_complex(self.feature_extract(evt).drop(
                self.buy_filter(BUY_EVT, ['buy', buy_item]), axis=0)))
        else:
            return self.add_tiles(self.join_complex(self.feature_extract(evt)), is_gamestate)

    def add_time_abd(self, frame, farmfails):
        time_abandoned = []
        for f in frame.itertuples():
            tx, ty, sess = f.tile['tx'], f.tile['ty'], f.sess_id
            fails = list(x for x in farmfails.itertuples() if
                         x.tile['tx'] == tx and x.tile['ty'] == ty and x.sess_id == sess)
            if len(fails) == 1:
                time_abandoned.append(fails[0].utc_time_secs - f.utc_time_secs)
            elif len(fails) > 1:
                time_abandoned.append(
                    max([fails[i].utc_time_secs - f.utc_time_secs for i in range(len(fails))]))
            else:
                time_abandoned.append(0)
        frame['time_abandoned'] = time_abandoned
        return frame

    def compose_results(self, frame):
        abd_ratio, avg_abd = {}, {}
        for sess in list(frame['sess_id'].drop_duplicates()):
            f_slice = list(frame[frame['sess_id'] == sess]['time_abandoned'].values)
            abd_ratio[sess] = sum([1 for x in f_slice if x > 0]) / (sum([1 for x in f_slice if x == 0]) or 1)
            avg_abd[sess] = sum(f_slice) / len(f_slice)

        frame['abd_ratio'] = [abd_ratio[x.sess_id] for x in frame.itertuples()]
        frame['avg_abd'] = [avg_abd[x.sess_id] for x in frame.itertuples()]
        if 'tile' not in frame.columns:
            return 'None abandoned' #might be something wrong here
        return frame[['sess_id', 'tile', 'utc_time_secs', 'time_abandoned', 'avg_abd', 'abd_ratio']]




## import standard libraries
import json
from typing import Any, Callable, Dict, Final, List, Optional
## import local files
from . import features
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.Feature import Feature
from ogd.games.ICECUBE.features import *
# from ogd.games.ICECUBE.DBExport import scene_map
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.Logger import Logger

## @class WaveExtractor
#  Extractor subclass for extracting features from Waves game data.

EXPORT_PATH : Final[str] = "games/ICECUBE/DBExport.json"
class IcecubeLoader(GeneratorLoader):

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule():
        return features
    
    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Optional[Feature]:
        ret_val : Optional[Feature] = None
        if extractor_params._count_index is None:
            match feature_type:
                case "ScenesEncountered":
                    ret_val = ScenesEncountered.ScenesEncountered(params=extractor_params)
                case "Session_Language":
                    ret_val = Session_Language.Session_Language(params=extractor_params)
                case "SessionDuration":
                    ret_val = SessionDuration.SessionDuration(params=extractor_params, session_id=self._session_id)
                case "HeadsetOnCount":
                    ret_val = HeadsetOnCount.HeadsetOnCount(params=extractor_params)
                case "ObjectSelectionsDuringVoiceover":
                    ret_val = ObjectSelectionsDuringVoiceover.ObjectSelectionsDuringVoiceover(params=extractor_params)
                case "SceneFailureCount":
                    ret_val = SceneFailureCount.SceneFailureCount(params=extractor_params)
                case _:
                    Logger.Log(f"'{feature_type}' is not a valid aggregate feature type for Icecube.")
        # Per-count features
        else:
            match feature_type:
                case "SceneFailures":
                    ret_val = SceneFailures.SceneFailures(params=extractor_params)
                case "TaskTimeToComplete":
                        ret_val = TaskTimeToComplete.TaskTimeToComplete(params=extractor_params)
                case "SceneDuration":
                        ret_val = SceneDuration.SceneDuration(params=extractor_params)
                case _:
                    Logger.Log(f"'{feature_type}' is not a valid aggregate feature for Icecube.")
            
        return ret_val

    def _loadDetector(self, detector_type:str, name:str, detector_args:Dict[str,Any], trigger_callback:Callable[[Event], None], count_index:Optional[int] = None) -> Optional[Detector]:
        Logger.Log(f"'{detector_type}' is not a valid detector for Waves.")
        return None

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the WaveLoader class.
    def __init__(self, player_id:str, session_id: str, game_schema:GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]=None):
        """Constructor for the WaveLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)
        self._scene_map = {"no-active-scene": 0}
        self._task_map = {}

        # Load Aqualab scenes export and map scene names to integer values
        # with open(EXPORT_PATH, "r") as file:
        #     export = json.load(file)

        #     task_num = 1
        #     for i, scene in enumerate(export["scenes"], start=1):
        #         self._scene_map[scene["id"]] = i
        #         self._diff_map[i] = scene["difficulties"]
        #         for task in scene["tasks"]:
        #             task_by_scene = scene["id"] + "_" + task["id"]
        #             self._task_map[task_by_scene] = task_num
        #             task_num += 1

        # Update level count
        # self._game_schema._max_level = len(self._scene_map) - 1


# import libraries
import logging
from typing import Any, List, Optional
from datetime import datetime, timedelta
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.utils.Logger import Logger

class ActivityDuration(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._session_id = None
        self._activity_start_time = None
        self._prev_timestamp = None
        self._time = 0
        self._activity_name = None
        self._activity_dict = dict()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["activity_begin", "activity_end"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return [] 

    def _updateFromEvent(self, event:Event) -> None:
        if event.SessionID != self._session_id:
            self._session_id = event.SessionID
            # if we jumped to a new session, we only want to count time up to last event, not the time between sessions.
            if self._activity_start_time and self._prev_timestamp:
                self._time += (self._prev_timestamp - self._activity_start_time).total_seconds()
                self._activity_start_time = event.Timestamp

        if event.EventName == "activity_begin":
            self._activity_start_time = event.Timestamp
            self._activity_name = event.event_data.get("activity_name")
            if not self._activity_name in self._activity_dict.keys():
                self._activity_dict[self._activity_name] = timedelta(0)
        elif event.EventName == "activity_end":
            if self._activity_start_time is not None:
                self._time += (event.Timestamp - self._activity_start_time).total_seconds()
                self._activity_dict[self._activity_name]+=timedelta(seconds=self._time)
                self._activity_start_time = None

        self._prev_timestamp = event.Timestamp



    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._activity_dict]


    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return [] 


# import standard libraries
from datetime import timedelta
from enum import IntEnum
from typing import Any, Dict, List, Union
# import locals
from ogd.core.utils.utils import ExportRow

class ResultStatus(IntEnum):
    NONE = 1
    SUCCESS = 2
    FAILURE = 3

class ExportResult:
    def __init__(self, columns:List[str] = [], values:List[ExportRow] = []):
        self._columns = columns
        self._values  = values

    @property
    def Columns(self) -> List[str]:
        return self._columns
    @Columns.setter
    def Columns(self, new_columns:List[str]):
        self._columns = new_columns

    @property
    def Values(self) -> List[ExportRow]:
        return self._values

    def ToDict(self) -> Dict[str, Union[List[str], List[ExportRow]]]:
        return {
            "cols":self._columns,
            "vals":self._values
        }

    def AppendRow(self, new_values:ExportRow):
        self._values.append(new_values)
    
    def ConcatRows(self, new_values:List[ExportRow]):
        self._values += new_values

class RequestResult:
    def __init__(self, msg:str="", status:ResultStatus=ResultStatus.NONE,
                 sess_ct : int = 0, duration:timedelta=timedelta()):
        self._message    = msg
        self._status     = status
        self._sess_ct    = sess_ct
        self._duration   = duration

    @property
    def Status(self) -> ResultStatus:
        return self._status

    @property
    def Message(self) -> str:
        return self._message

    @property
    def SessionCount(self):
        return self._sess_ct
    @SessionCount.setter
    def SessionCount(self, new_ct:int):
        self._sess_ct = new_ct

    @property
    def Duration(self) -> timedelta:
        return self._duration
    @Duration.setter
    def Duration(self, new_duration):
        self._duration = new_duration

    def RequestSucceeded(self, msg:str):
        self._status = ResultStatus.SUCCESS
        self._message = msg

    def RequestErrored(self, msg:str):
        self._status = ResultStatus.FAILURE
        self._message = msg



# import libraries
import logging
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.games.AQUALAB.features.PerJobFeature import PerJobFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class JobDiveSitesCount(PerJobFeature):
    def __init__(self, params:GeneratorParameters, job_map:dict):
        super().__init__(params=params, job_map=job_map)
        self._count = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["begin_dive"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        self._count += 1

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._count]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



# import libraries
import logging
from datetime import timedelta
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.games.AQUALAB.features.PerJobFeature import PerJobFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class JobActiveTime(PerJobFeature):

    def __init__(self, params:GeneratorParameters, job_map:dict):
        super().__init__(params=params, job_map=job_map)
        self._total_time = timedelta(0)
        if self.ExtractionMode == ExtractionMode.PLAYER:
            self._session_id      = None
            self._last_start_time = None
            self._last_event_time = None
            self._last_event_session = None
            self._last_event_name = None
            self._last_event_index = None
            self._last_event_session = None
            self._last_event_name = None
            self._last_event_index = None
        else:
            raise NotImplementedError(f"Got invalid export mode of {self.ExtractionMode.name} in JobActiveTime!")

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]
        # if self.ExtractionMode == ExtractionMode.PLAYER \
        # or self.ExtractionMode == ExtractionMode.SESSION:
        #     return ["all_events"]
        # else:
        #     return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []
        # if self.ExtractionMode == ExtractionMode.POPULATION:
        #     return ["JobActiveTime"]
        # else:
        #     return []

    def _updateFromEvent(self, event:Event) -> None:
        self._player_id = event.UserID
        if event.SessionID != self._session_id:
            self._handleNewSession(event)

        if event.EventName == "accept_job":
            self._last_start_time = event.timestamp
        elif event.EventName == "switch_job":
            new_job = event.GameState.get('job_name', event.EventData.get('job_name', None))
            if new_job is None:
                raise KeyError("Could not find key 'job_name' in GameState or EventData!")
            old_job = event.EventData["prev_job_name"]
            # if we switched into "this" job, this becomes new start time
            if self._job_map.get(new_job, None) == self.CountIndex:
                self._last_start_time = event.Timestamp
                if event.Timestamp is None:
                    self.WarningMessage(f"JobActiveTime received a switch_job event with Timestamp == None!")
            # if we switched out of "this" job, update total time in the job.
            # note, if "this" job is no-active-job, we don't care. Further, if we switched into no-active-job, then we just completed a job and don't care.
            elif self._job_map.get(old_job, None) == self.CountIndex and new_job != "no-active-job" and old_job != "no-active-job":
                self._updateTotalTime()
        elif event.EventName == "complete_job":
            self._updateTotalTime()
        # if we got an event unrelated to job change, but didn't have a start (e.g. new session started mid-job), then make now the start.
        elif self._last_start_time is None:
            self._last_start_time = event.Timestamp
        # if we got an event earlier than last start, we are out of order
        elif self._last_start_time > event.Timestamp:
            self.WarningMessage(f"Got out-of-order events in JobActiveTime; event {event.EventName}:{event.EventSequenceIndex} had timestamp {event.Timestamp} earlier than start event, with time {self._last_start_time}!")
            self._last_start_time = event.Timestamp

        # finally, update latest timestamp
        if self._last_event_time and self._last_event_time > event.Timestamp:
            self.WarningMessage(f"Got out-of-order events in SessionDuration; event {event.EventName}:{event.EventSequenceIndex} had timestamp {event.Timestamp} earlier than end event, with time {self._last_event_time}!")
        else:
            self._last_event_time = event.timestamp
            self._last_event_session = event.SessionID
            self._last_event_name = event.UserID
            self._last_event_index = event.EventSequenceIndex

    def _updateFromFeatureData(self, feature:FeatureData):
        return
        # if self.ExtractionMode == ExtractionMode.PLAYER \
        # or self.ExtractionMode == ExtractionMode.SESSION:
        #     pass
        # elif self.ExtractionMode == ExtractionMode.POPULATION:
        #     self._handle_population(feature=feature)

    def _getFeatureValues(self) -> List[Any]:
        return [self._total_time.total_seconds()]

    def _validateEventCountIndex(self, event:Event):
        ret_val : bool = False

        _current_job = event.GameState.get('job_name', event.EventData.get('job_name', None))
        if _current_job is None:
            raise KeyError("Could not find key 'job_name' in GameState or EventData!")
        if self._job_map.get(_current_job, None) is not None:
            if self._job_map.get(_current_job, None) == self.CountIndex:
                    ret_val = True
            elif event.EventName == "switch_job":
                # if we got switch job, and were switching away from this instance's job, we still want to process it.
                old_job = event.EventData["prev_job_name"]
                if self._job_map.get(old_job, None) == self.CountIndex:
                    ret_val = True
        else:
            self.WarningMessage(f"Got invalid job_name data in {type(self).__name__}")

        return ret_val


    # *** Optionally override public functions. ***

    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"

    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.PLAYER]

    # *** PRIVATE METHODS ***

    def _handleNewSession(self, event:Event):
            _old_sess = self._session_id
            self._session_id = event.SessionID
            # if we jumped to a new session, we only want to count time up to last event, then skip the time between sessions to new timestamp;
            # but only if we had a previous session, i.e. this isn't the first event seen.
            if _old_sess is not None:
                # Logger.Log(f"JobActiveTime attempting to update total time for {event.UserID} ({_old_sess} -> {self._session_id}) following change in session, index={event.EventSequenceIndex}", logging.INFO)
                self._updateTotalTime()
                # Logger.Log("Done", logging.INFO)
                if event.app_version == 'Aqualab' or event.app_version == 'None':
                    current_job = event.EventData.get("job_name", {}).get('string_value')
                else:
                    current_job = event.EventData.get("job_name")
                if current_job is not None and self._job_map.get(current_job, None) == self.CountIndex:
                    self._last_start_time = event.Timestamp
                    self._last_event_time = event.Timestamp
                if event.Timestamp is None:
                    Logger.Log(f"JobActiveTime received an initial event with Timestamp == None!", logging.WARN)
            # _old_sess = self._session_id
            # self._session_id = event.SessionID
            # # if we jumped to a new session, we only want to count time up to last event, then skip the time between sessions to new timestamp;
            # # but only if we had a previous session, i.e. this isn't the first event seen.
            # if _old_sess is not None:
            #     # Logger.Log(f"JobActiveTime attempting to update total time for {event.UserID} ({_old_sess} -> {self._session_id}) following change in session, index={event.EventSequenceIndex}", logging.INFO)
            #     self._updateTotalTime()
            #     # Logger.Log("Done", logging.INFO)
            #     self._last_start_time = event.timestamp
            #     if event.Timestamp is None:
            #         self.WarningMessage(f"JobActiveTime received an initial event with Timestamp == None!")

    def _updateTotalTime(self):
        if self._last_start_time:
            if self._last_event_time:
                self._total_time += (self._last_event_time - self._last_start_time)
                self._last_start_time = None
                self._last_event_time = None
            else:
                self.WarningMessage(f"JobActiveTime could not update total time, missing previous event time!")
        elif self.ExtractionMode == ExtractionMode.PLAYER:
            self.WarningMessage(f"JobActiveTime could not update total time for player {self._player_id}, session {self._session_id} missing start time!")

    # def _handle_population(self, feature:FeatureData):
    #     if feature.ExportMode == ExtractionMode.PLAYER:
    #         _val = feature.FeatureValues[0]
    #         if type(_val) == timedelta:
    #             self._total_seconds += _val
    #         else:
    #             Logger.Log(f"JobActiveTime for population got invalid value {_val} of type {type(_val)} for column {feature.FeatureNames[0]}", logging.WARN)
    #     else:
    #         Logger.Log(f"JobActiveTime for population got feature data for mode {feature.ExportMode.name}", logging.DEBUG)



# import libraries
import json
from datetime import datetime, timedelta
from os import environ
from sqlite3 import Timestamp
from typing import Any, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
# import local files
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.games.JOWILDER import Jowilder_Enumerators as je
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

# NOTE: Assumptions are: 1. All click events occured in the order like xxxx111xx222x1x3. 2. Use "text_fqid" to identify interactions. 3. The first interaction "tunic.historicalsociety.closet.intro" makes no sense so we don't need to consider it. That is, there are 190 interactions in total, but we only count 189. And we should confirm that, this tunic.historicalsociety.closet.intro doesn't occur anywhere else.


class ClickTrack:
    def __init__(self, start_time:Optional[datetime] = None, game_start: bool = False, this_click:Optional[Event]=None, last_click:Optional[Event]=None) -> None:
        self._game_start : bool = game_start
        self._game_start_time : Optional[datetime] = start_time
        self._last_click : Optional[Event] = last_click
        self._this_click : Optional[Event] = this_click
        self._time_between : Optional[timedelta] = None
        # 1 means searching in an interaction; 0 means searching in an non-interaction click.
        self._search_state : int = 0

    # TODO: Add more property decorators fuction
    @property
    def LastClickTime(self) -> Optional[datetime]:
        if self._last_click is not None:
            return self._last_click.Timestamp
        else:
            return None

    @property
    def LastInteractionIndex(self) -> Optional[int]:
        if self._last_click is not None:
            _interaction = self._last_click.EventData.get("text_fqid") \
                        or self._last_click.EventData.get("cur_cmd_fqid") \
                        or "FQID NOT FOUND"
            return je.fqid_to_enum.get(_interaction)
        else:
            return None

    @property
    def GameStart(self) -> bool:
        return self._game_start

    @property
    def GameStartTime(self) -> Optional[datetime]:
        return self._game_start_time

    @staticmethod
    def EventEq(event1:Event, event2:Event) -> bool:
        if event1.SessionID == event2.SessionID and event1.EventSequenceIndex == event2.EventSequenceIndex:
            return True
        return False

    def Update(self, event:Event) -> timedelta:
        # event_sequence_index
        ret_val : timedelta = timedelta(0)
        if self._this_click is not None and ClickTrack.EventEq(self._this_click, event):
            ret_val = self._time_between or ret_val # if self._time_between was non-null, return it.
        else:
            self._last_click = self._this_click
            self._this_click = event
            if self._last_click is not None:
                self._time_between = self._this_click.Timestamp - self._last_click.Timestamp
                ret_val = self._time_between
        return ret_val
    
    def StartNewInteraction(self, this_interaction : Optional[int]):
        if this_interaction is not None and (self._search_state == 0 or this_interaction != self.LastInteractionIndex):
            return True
    
    def startGame(self, event: Event):
        self._game_start = True
        self._game_start_time = event.Timestamp
        self._this_click = event
        self._last_click = event


clicks_track = ClickTrack()

class Interaction(PerCountFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self._interaction : Optional[int] = None
        self._interaction_time : timedelta = timedelta(0)
        self._first_encounter_time : timedelta = timedelta(0)
        self._num_encounters : int = 0
        self._to : timedelta = timedelta(0)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _validateEventCountIndex(self, event: Event):
        if event.EventName == "CUSTOM.1":
            return True
        _fqid = event.EventData.get("text_fqid") or event.EventData.get("cur_cmd_fqid") or "FQID NOT FOUND"
        self._interaction = je.fqid_to_enum.get(_fqid)
        if self._interaction is None:
            return self.CountIndex == clicks_track.LastInteractionIndex
        else:
            return self._interaction == self.CountIndex

        
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        # NOTE: Count all the click events
        return [f"CUSTOM.{i}" for i in range(3,12)] + ["CUSTOM.1"]
        # CUSTOM.X, X in [3,12) = ['navigate_click','notebook_click', 'map_click', 'notification_click', 'object_click', 'observation_click', 'person_click', 'cutscene_click', 'wildcard_click']

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return []

    def _updateFromEvent(self, event: Event) -> None:
        if not clicks_track.GameStart:
            if event.EventName == "CUSTOM.1":
                clicks_track.startGame(event)
                return
            else:
                raise(ValueError("A startgame event needed!"))
        elif event.EventName == "CUSTOM.1": 
            if clicks_track._this_click is not None and clicks_track.EventEq(event, clicks_track._this_click):
                return
            # else:
            #     raise(ValueError("Too many startgame events!"))

        if event.EventName == "CUSTOM.11" and event.EventData.get("cur_cmd_type") == 2:
            return
        if self._interaction is None:
            clicks_track.Update(event)
            clicks_track._search_state = 0
            return
        
        clicks_track.Update(event)
        if clicks_track.LastClickTime is not None:
            self._interaction_time += event.Timestamp - clicks_track.LastClickTime
            if clicks_track.StartNewInteraction(self._interaction):
                clicks_track._search_state = 1
                if self._num_encounters == 0:
                    if clicks_track.GameStartTime is not None:
                        # The interaction starts from last click but recorded from this click
                        self._to = clicks_track.LastClickTime - clicks_track.GameStartTime
                    else:
                        raise ValueError(f"Interaction feature tried to get _to value, but clicks_track didn't have a game start time.")
                self._num_encounters += 1
            if self._num_encounters <= 1:
                self._first_encounter_time = self._interaction_time
        else:
            raise ValueError(f"Interaction feature tried to get interaction time, but clicks_track didn't have a previous click.")
        
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._interaction_time, self._first_encounter_time, self._num_encounters, self._to]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["FirstEncounterTime", "NumEncounter", "TimeTo"]

    def BaseFeatureSuffix(self) -> str:
        return "Time"



"""# import libraries
from typing import Any, Dict, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class CountyBuildCount(Feature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.build_counts: Dict[str, int] = {}

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["execute_build_queue"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        county_name = event.EventData.get("county_name", None)
        if county_name:
            if county_name not in self.build_counts:
                self.build_counts[county_name] = 1
            else:
                self.build_counts[county_name] += 1

    def _updateFromFeatureData(self, feature: FeatureData):
        pass

    def _getFeatureValues(self) -> List[Any]:
        return [self.build_counts]

    def Subfeatures(self) -> List[str]:
        return []
"""

"""# import libraries
from typing import Any, Dict, List
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

# import PerCountyFeature
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.games.BLOOM.features.PerCountyFeature import PerCountyFeature

class CountyBuildCount(PerCountyFeature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.build_counts: Dict[str, Dict[str, int]] = {county: 0 for county in self.COUNTY_LIST}

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["execute_build_queue"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _validateEventCountIndex(self, event: Event):
        ret_val: bool = False

        county_name = event.EventData.get('county_name', "COUNTY NAME NOT FOUND")
        if county_name is not None:
            if county_name in self._county_map and self._county_map[county_name] == self.CountIndex:
                ret_val = True
        else:
            self.WarningMessage(f"Got invalid county_name data in {type(self).__name__}")

        return ret_val

    def _updateFromEvent(self, event: Event) -> None:
        county_name = event.EventData.get("county_name", None)
        if county_name and county_name in self.build_counts:
            self.build_counts[county_name] += 1

    def _updateFromFeatureData(self, feature: FeatureData):
        pass

    def _getFeatureValues(self) -> List[Any]:
        return [self.build_counts]

    def Subfeatures(self) -> List[str]:
        return []"""




# import libraries
from typing import Any, Dict, List
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

# import PerCountyFeature
from ogd.core.generators.extractors.Feature import Feature
from ogd.games.BLOOM.features.PerCountyFeature import PerCountyFeature

class CountyBuildCount(PerCountyFeature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.count : int = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["click_execute_build"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []


    def _updateFromEvent(self, event: Event) -> None:
        self.count += 1

    def _updateFromFeatureData(self, feature: FeatureData):
        pass

    def _getFeatureValues(self) -> List[Any]:
        return [self.count]

    def Subfeatures(self) -> List[str]:
        return []



# import libraries
import abc
import logging
from importlib import import_module
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Type
# import locals
from ogd.core.generators.extractors import builtin
from ogd.core.generators.extractors.builtin.BuiltinExtractor import BuiltinExtractor
from ogd.core.generators.extractors.builtin import *
from ogd.core.generators.Generator import Generator, GeneratorParameters
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.extractors.Extractor import Extractor
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.Logger import Logger

class GeneratorLoader(abc.ABC):

    # *** ABSTRACTS ***
    
    @abc.abstractmethod
    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Optional[Extractor]:
        pass
    
    @abc.abstractmethod
    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Optional[Detector]:
        pass

    @staticmethod
    @abc.abstractmethod
    def _getFeaturesModule() -> ModuleType:
        pass

    # *** BUILT-INS & PROPERTIES ***

    _derived_builtins : Dict[str, Type[BuiltinExtractor]] = {}
    def __init__(self, player_id:str, session_id:str, game_schema:GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        """Base constructor for Loader classes.
        Sets the player and session identifiers, as well as the mode in which to load generators.
        Also sets the schema, which contains config info, and any overrides.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id of the session from which we will extract features.
        :type session_id: str
        :param game_schema: A dictionary that defines how the game data itself is structured.
        :type game_schema: GameSchema
        :param mode: _description_
        :type mode: ExtractionMode
        :param feature_overrides: _description_
        :type feature_overrides: Optional[List[str]]
        """
        # TODO : seems like Loader shouldn't really need player ID,  session ID, or overrides, consider removing.
        self._player_id   : str            = player_id
        self._session_id  : str            = session_id
        self._game_schema : GameSchema     = game_schema
        self._mode        : ExtractionMode = mode
        self._overrides   : Optional[List[str]] = feature_overrides

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***
    
    def LoadDetector(self, detector_type:str, name:str, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None], count_index:Optional[int] = None) -> Optional[Detector]:
        ret_val = None

        params = GeneratorParameters(name=name, description=schema_args.get('description',""), mode=self._mode, count_index=count_index)
        ret_val = self._loadDetector(detector_type=detector_type, extractor_params=params, schema_args=schema_args, trigger_callback=trigger_callback) \
               or self._loadBuiltinDetector(detector_type=detector_type, extractor_params=params, schema_args=schema_args, trigger_callback=trigger_callback)

        return ret_val

    def LoadFeature(self, feature_type:str, name:str, schema_args:Dict[str,Any], count_index:Optional[int] = None) -> Optional[Extractor]:
        ret_val = None

        if self._validateMode(feature_type=feature_type):
            params = GeneratorParameters(name=name, description=schema_args.get('description',""), mode=self._mode, count_index=count_index)
            ret_val = self._loadFeature(feature_type=feature_type, extractor_params=params, schema_args=schema_args) \
                   or self._loadBuiltinFeature(feature_type=feature_type, extractor_params=params, schema_args=schema_args)

        return ret_val

    def GetFeatureClass(self, feature_type:str) -> Optional[Type[Extractor]]:
        ret_val : Optional[Type[Extractor]] = None
        game_module = self._getFeaturesModule()
        try:
            # Try to find feature module in the game module, falling back on finding it in builtin module.
            feature_mod = getattr(game_module, feature_type, None) or getattr(builtin, feature_type)
            # if feature_mod is None:
            #     Logger.Log(f"Didn't find module for feature {feature_type} in module {game_module}, searching in builtin module instead.")
            #     feature_mod = 
                # if feature_mod is None:
                #     Logger.Log(f"Didn't find module for feature {feature_type} in builtin module.")
        except NameError as err:
            Logger.Log(f"Could not find class {feature_type} in module `{game_module}` or in `builtin`, a NameError occurred:\n{err}\nSearching `builtin` module instead.", logging.WARN)
            raise err
        else:
            # If the above did not generate an error, then try to get actual feature class from its module.
            # If not found, then we fall back on getting builtin directly, though in practice this is probably redundant.
            ret_val = getattr(feature_mod, feature_type, None) or self.GetBuiltinFeatureClass(feature_type=feature_type)
            # if ret_val is None:
            #     ret_val = 
        finally:
            return ret_val

    def GetBuiltinFeatureClass(self, feature_type:str) -> Optional[Type[BuiltinExtractor]]:
        ret_val : Optional[Type[BuiltinExtractor]] = None
        try:
            feature_mod = getattr(builtin, feature_type)
            ret_val = getattr(feature_mod, feature_type)
        except NameError as err:
            Logger.Log(f"Could not find class {feature_type} in module game module or `builtin`, a NameError occurred:\n{err}", logging.WARN)
        finally:
            return ret_val

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    def _validateMode(self, feature_type) -> bool:
        ret_val = False

        feature_class : Optional[Type[Extractor]]  = self.GetFeatureClass(feature_type=feature_type)
        if feature_class is None:
            Logger.Log(f"In GeneratorLoader, skipping feature class `{feature_type}`, which could not be found.", logging.WARN)
        else:
            ret_val = self._mode in feature_class.AvailableModes()

        return ret_val

    def _loadBuiltinFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Extractor:
        ret_val : BuiltinExtractor
        if extractor_params._name not in GeneratorLoader._derived_builtins.keys():
            feature_class = self.GetBuiltinFeatureClass(feature_type=feature_type)
            if feature_class is not None:
                GeneratorLoader._derived_builtins[extractor_params._name] = feature_class._createDerivedGenerator(params=extractor_params, schema_args=schema_args)
        # Session-level features.
        if extractor_params._name in GeneratorLoader._derived_builtins.keys():
            ret_val = GeneratorLoader._derived_builtins[extractor_params._name](params=extractor_params, schema_args=schema_args)
        else:
            raise NotImplementedError(f"'{feature_type}' is not a valid built-in feature.")
        return ret_val

    def _loadBuiltinDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Detector:
        raise NotImplementedError(f"'{detector_type}' is not a valid built-in detector.")



# import libraries
import json
from typing import Any, List, Optional
# import local files
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.games.ICECUBE.DBExport import scene_map
from ogd.games.ICECUBE.features.PerSceneFeature import PerSceneFeature
from ogd.games.ICECUBE.DBExport import scenes_list
from ogd.core.utils.Logger import Logger

# import libraries
import logging

scenes_map = {"ICE":0, "VOYAGER":1, "NOTHING":2, "EXTREME":3, "EARTH":4, "CREDITS":5, "None":6}

class SceneFailures(PerCountFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._failure_count = 0
        self._name = None
        self._found = False
        self._task_idx = 0
        self._idx = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["scene_begin","scene_end", "failed"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        # Logger.Log(f"event data is {event.EventData}")
        if event.EventName == "failed":
            self._failure_count += 1
        
        
        

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._failure_count]

    # *** Optionally override public functions. ***
    def _validateEventCountIndex(self, event: Event):
        ret_val : bool = False

        # scene_data = event.EventData.get("scene_name")
        scene_data = event.game_state.get("scene_name")
        # Logger.Log(f"scene_data collected = {scene_data}")
        
        # if scene_data is None:
        #     scene_data = "None"
        #     ret_val = True
        
        if scene_data is not None:
            # Logger.Log(f"session_id is {event.SessionID}")
            # Logger.Log(f"event name is {event.event_name}")
            if scenes_map[scene_data] == self.CountIndex:
                # Logger.Log(f"count index is {self.CountIndex}")
                # Logger.Log(f"map index is {scenes_map[scene_data]}")
            # Logger.Log(f"scene_data is {scene_data}")
                ret_val = True
        
        # else:
        #     self.WarningMessage(f"Got invalid job_name data in {type(self).__name__}")

        return ret_val
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "2"